package main

import (
	"flag"
	"fmt"
	"log"
	"os"
	"runtime/pprof"
	"syscall"
	"time"
)

//             12345678
const Niter = 1000000

// xpread is pread(2) but aborts on an error or when fewer bytes was read
func xpread(fd int, buf []byte, offset int64) {
	n, err := syscall.Pread(fd, buf, offset)
	if err != nil {
		panic(err)
	}
	if n != len(buf) {
		panic(fmt.Errorf("pread(%v) -> %v", len(buf), n))
	}
}

// xreadat is ReadAt but aborts on an error or when fewer bytes was read
func xreadat(f *os.File, buf []byte, offset int64) {
	n, err := f.ReadAt(buf, offset)
	if err != nil {
		panic(err)
	}
	if n != len(buf) {
		panic(fmt.Errorf("readat(%v) -> %v", len(buf), n))
	}
}


const BufSize = 4096
var buf [BufSize]byte

var cpuprofile = flag.String("cpuprofile", "", "write cpu profile `file`")

func main() {
	flag.Parse()
	if len(flag.Args()) < 1 {
		log.Fatalf("Usage: %s <file>", os.Args[0])
	}

	if *cpuprofile != "" {
		f, err := os.Create(*cpuprofile)
		if err != nil {
			log.Fatal("could not create CPU profile: ", err)
		}
		if err := pprof.StartCPUProfile(f); err != nil {
			log.Fatal("could not start CPU profile: ", err)
		}
		defer pprof.StopCPUProfile()
	}


	f, err := os.Open(flag.Args()[0])
	if err != nil {
		log.Fatal(err)
	}
	fd := int(f.Fd())

	for size := 0; size <= BufSize; {
		sbuf := buf[:size]

		// warmup
		xpread(fd, sbuf, /*offset=*/0)

		Tstart := time.Now()

		for i := 0; i < Niter; i++ {
			xpread(fd, sbuf, /*offset=*/0)
		}

		Tend := time.Now()

		fmt.Printf("%v\t%.1f\n", size, float64(Tend.Sub(Tstart).Nanoseconds()) / Niter)

		switch {
		case size < 16:
			size += 1	// detailed start
		case size < 128:
			size += 16
		default:
			size += 128
		}
	}
}
